"""
Schedule module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import uuid
from typing import Union, List


class Schedule:
  """
  Schedule class
  """

  def __init__(self):
    self._id = None
    self._type = None
    self._values = None
    self._data_type = None
    self._time_step = None
    self._time_range = None
    self._day_types = None

  @property
  def id(self):
    """
    Get schedule id, a universally unique identifier randomly generated
    :return: str
    """
    if self._id is None:
      self._id = uuid.uuid4()
    return self._id

  @property
  def type(self) -> Union[None, str]:
    """
    Get schedule type
    :return: None or str
    """
    return self._type

  @type.setter
  def type(self, value):
    """
    Set schedule type
    :param: str
    """
    if value is not None:
      self._type = str(value)

  @property
  def values(self):
    """
    Get schedule values
    :return: [Any]
    """
    return self._values

  @values.setter
  def values(self, value):
    """
    Set schedule values
    :param: [Any]
    """
    self._values = value

  @property
  def data_type(self) -> Union[None, str]:
    """
    Get schedule data type from:
    ['any_number', 'fraction', 'on_off', 'temperature', 'humidity', 'control_type']
    :return: None or str
    """
    return self._data_type

  @data_type.setter
  def data_type(self, value):
    """
    Set schedule data type
    :param: str
    """
    if value is not None:
      self._data_type = str(value)

  @property
  def time_step(self) -> Union[None, str]:
    """
    Get schedule time step from:
    ['second', 'minute', 'hour', 'day', 'week', 'month']
    :return: None or str
    """
    return self._time_step

  @time_step.setter
  def time_step(self, value):
    """
    Set schedule time step
    :param: str
    """
    if value is not None:
      self._time_step = str(value)

  @property
  def time_range(self) -> Union[None, str]:
    """
    Get schedule time range from:
    ['minute', 'hour', 'day', 'week', 'month', 'year']
    :return: None or str
    """
    return self._time_range

  @time_range.setter
  def time_range(self, value):
    """
    Set schedule time range
    :param: str
    """
    if value is not None:
      self._time_range = str(value)

  @property
  def day_types(self) -> Union[None, List[str]]:
    """
    Get schedule day types, as many as needed from:
    ['monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday', 'holiday', 'winter_design_day',
    'summer_design_day']
    :return: None or [str]
    """
    return self._day_types

  @day_types.setter
  def day_types(self, value):
    """
    Set schedule day types
    :param: [str]
    """
    if value is not None:
      self._day_types = [str(i) for i in value]



"""
Construction catalog factory, publish the construction information
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from pathlib import Path
from typing import TypeVar

from hub.catalog_factories.construction.nrcan_catalog import NrcanCatalog
from hub.catalog_factories.construction.nrel_catalog import NrelCatalog
from hub.catalog_factories.construction.eilat_catalog import EilatCatalog
from hub.helpers.utils import validate_import_export_type

Catalog = TypeVar('Catalog')


class ConstructionCatalogFactory:
  """
  Construction catalog factory class
  """
  def __init__(self, handler, base_path=None):
    if base_path is None:
      base_path = Path(Path(__file__).parent.parent / 'data/construction')
    self._handler = '_' + handler.lower()
    validate_import_export_type(ConstructionCatalogFactory, handler)
    self._path = base_path

  @property
  def _nrel(self):
    """
    Retrieve NREL catalog
    """
    return NrelCatalog(self._path)

  @property
  def _nrcan(self):
    """
    Retrieve NRCAN catalog
    """
    return NrcanCatalog(self._path)

  @property
  def _eilat(self):
    """
    Retrieve Eilat catalog
    """
    return EilatCatalog(self._path)

  @property
  def catalog(self) -> Catalog:
    """
    Enrich the city given to the class using the class given handler
    :return: Catalog
    """
    return getattr(self, self._handler, lambda: None)



"""
Test db factory
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
import distutils.spawn
import glob
import json
import logging
import os
import subprocess
import unittest
from pathlib import Path
from unittest import TestCase

import sqlalchemy.exc
from sqlalchemy import create_engine
from sqlalchemy.exc import ProgrammingError

import hub.helpers.constants as cte
from hub.exports.energy_building_exports_factory import EnergyBuildingsExportsFactory
from hub.exports.exports_factory import ExportsFactory
from hub.helpers.data.montreal_function_to_hub_function import MontrealFunctionToHubFunction
from hub.imports.construction_factory import ConstructionFactory
from hub.imports.energy_systems_factory import EnergySystemsFactory
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.results_factory import ResultFactory
from hub.imports.usage_factory import UsageFactory
from hub.imports.weather_factory import WeatherFactory
from hub.persistence.db_control import DBControl
from hub.persistence.models import City, Application, CityObject, SimulationResults
from hub.persistence.models import User, UserRoles
from hub.persistence.repository import Repository


class Control:
  _skip_test = False
  _skip_reason = 'PostgreSQL not properly installed in host machine'

  def __init__(self):
    """
    Test
    setup
    :return: None
    """
    self._skip_test = False
    # Create test database
    dotenv_path = Path("{}/.local/etc/hub/.env".format(os.path.expanduser('~'))).resolve()
    if not dotenv_path.exists():
      self._skip_test = True
      self._skip_reason = f'.env file missing at {dotenv_path}'
      return
    dotenv_path = str(dotenv_path)
    repository = Repository(db_name='montreal_retrofit_test', app_env='TEST', dotenv_path=dotenv_path)
    engine = create_engine(repository.configuration.connection_string)
    try:
      # delete test database if it exists
      connection = engine.connect()
      connection.close()
    except ProgrammingError:
      logging.info('Database does not exist. Nothing to delete')
    except sqlalchemy.exc.OperationalError as operational_error:
      self._skip_test = True
      self._skip_reason = f'{operational_error}'
      return

    self._database = DBControl(
      db_name=repository.configuration.db_name,
      app_env='TEST',
      dotenv_path=dotenv_path)

    self._application_uuid = '60b7fc1b-f389-4254-9ffd-22a4cf32c7a3'
    self._application_id = 1
    self._user_id = 1
    self._pickle_path = 'tests_data/pickle_path.bz2'

  @property
  def database(self):
    return self._database

  @property
  def application_uuid(self):
    return self._application_uuid

  @property
  def application_id(self):
    return self._application_id

  @property
  def user_id(self):
    return self._user_id

  @property
  def skip_test(self):
    return self._skip_test

  @property
  def insel(self):
    return distutils.spawn.find_executable('insel')

  @property
  def sra(self):
    return distutils.spawn.find_executable('sra')

  @property
  def skip_insel_test(self):
    return self.insel is None

  @property
  def skip_reason(self):
    return self._skip_reason

  @property
  def message(self):
    return self._skip_reason

  @property
  def pickle_path(self):
    return self._pickle_path


control = Control()


class TestDBFactory(TestCase):
  """
TestDBFactory
"""

  @unittest.skipIf(control.skip_test, control.skip_reason)
  def test_retrieve_results(self):
    request_values = {
      "scenarios": [
        {
          "current status": ["01002777", "01002773", "01036804"]
        },
        {
          "skin retrofit": ["01002777", "01002773", "01036804"]
        },
        {
          "system retrofit and pv": ["01002777", "01002773", "01036804"]
        },
        {
          "skin and system retrofit with pv": ["01002777", "01002773", "01036804"]
        }


      ]
    }
    results = control.database.results(control.user_id, control.application_id, request_values)
    print(results)



"""
Dictionaries module for Eilat function to hub function
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import hub.helpers.constants as cte


class EilatFunctionToHubFunction:
  """
  Eilat function to hub function class
  """

  def __init__(self):
    self._dictionary = {'Residential': cte.RESIDENTIAL,
                        'Dormitory': cte.DORMITORY,
                        'Hotel employ': cte.HOTEL
                        }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary



"""
Cost catalog publish the life cycle cost
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Atiya atiya.atiya@mail.concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from pathlib import Path
from typing import TypeVar
from hub.catalog_factories.cost.montreal_custom_catalog import MontrealCustomCatalog

Catalog = TypeVar('Catalog')


class CostsCatalogFactory:
  """
  CostsCatalogFactory class
  """
  def __init__(self, file_type, base_path=None):
    if base_path is None:
      base_path = Path(Path(__file__).parent.parent / 'data/costs')
    self._catalog_type = '_' + file_type.lower()
    self._path = base_path

  @property
  def _montreal_custom(self):
    """
    Retrieve Montreal Custom catalog
    """
    return MontrealCustomCatalog(self._path)

  @property
  def catalog(self) -> Catalog:
    """
    Return a cost catalog
    :return: CostCatalog
    """
    return getattr(self, self._catalog_type, lambda: None)



"""
EnergyBuildingsExportsFactory exports a city into several formats related to energy in buildings
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de uribarri pilar.monsalvete@concordia.ca
"""

from pathlib import Path

import requests

from hub.exports.building_energy.energy_ade import EnergyAde
from hub.exports.building_energy.idf import Idf
from hub.exports.building_energy.insel.insel_monthly_energy_balance import InselMonthlyEnergyBalance
from hub.helpers.utils import validate_import_export_type
from hub.imports.weather.helpers.weather import Weather as wh


class EnergyBuildingsExportsFactory:
  """
  Energy Buildings exports factory class
  """
  def __init__(self, handler, city, path, custom_insel_block='d18599', target_buildings=None):
    self._city = city
    self._export_type = '_' + handler.lower()
    validate_import_export_type(EnergyBuildingsExportsFactory, handler)
    if isinstance(path, str):
      path = Path(path)
    self._path = path
    self._custom_insel_block = custom_insel_block
    self._target_buildings = target_buildings

  @property
  def _energy_ade(self):
    """
    Export to citygml with application domain extensions
    :return: None
    """
    return EnergyAde(self._city, self._path)

  @property
  def _idf(self):
    """
    Export the city to Energy+ idf format

    When target_buildings is set, only those will be calculated and their energy consumption output, non-adjacent
    buildings will be considered shading objects and adjacent buildings will be considered adiabatic.

    Adjacent buildings are provided they will be considered heated so energy plus calculations are more precise but
    no results will be calculated to speed up the calculation process.

    :return: None
    """
    idf_data_path = (Path(__file__).parent / './building_energy/idf_files/').resolve()
    url = wh().epw_file(self._city.region_code)
    weather_path = (Path(__file__).parent.parent / f'data/weather/epw/{url.rsplit("/", 1)[1]}').resolve()
    if not weather_path.exists():
      with open(weather_path, 'wb') as epw_file:
        epw_file.write(requests.get(url, allow_redirects=True).content)
    return Idf(self._city, self._path, (idf_data_path / 'Minimal.idf'), (idf_data_path / 'Energy+.idd'), weather_path,
               target_buildings=self._target_buildings)

  @property
  def _insel_monthly_energy_balance(self):
    """
    Export to Insel MonthlyEnergyBalance
    :return: None
    """
    return InselMonthlyEnergyBalance(self._city, self._path, self._custom_insel_block)

  def export(self):
    """
    Export the city given to the class using the given export type handler
    :return: None
    """
    return getattr(self, self._export_type, lambda: None)



"""
Domestic Hot Water module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from typing import Union, List
from hub.city_model_structure.attributes.schedule import Schedule


class DomesticHotWater:
  """
  DomesticHotWater class
  """
  def __init__(self):
    self._density = None
    self._peak_flow = None
    self._service_temperature = None
    self._schedules = None

  @property
  def density(self) -> Union[None, float]:
    """
    Get domestic hot water load density in Watts per m2
    :return: None or float
    """
    return self._density

  @density.setter
  def density(self, value):
    """
    Set domestic hot water load density in Watts per m2
    :param value: float
    """
    if value is not None:
      self._density = float(value)

  @property
  def peak_flow(self) -> Union[None, float]:
    """
    Get domestic hot water peak_flow density in m3 per second and m2
    :return: None or float
    """
    return self._peak_flow

  @peak_flow.setter
  def peak_flow(self, value):
    """
    Set domestic hot water peak_flow density in m3 per second and m2
    :return: None or float
    """
    self._peak_flow = value

  @property
  def service_temperature(self) -> Union[None, float]:
    """
    Get service temperature in degrees Celsius
    :return: None or float
    """
    return self._service_temperature

  @service_temperature.setter
  def service_temperature(self, value):
    """
    Set service temperature in degrees Celsius
    :param value: float
    """
    if value is not None:
      self._service_temperature = float(value)

  @property
  def schedules(self) -> Union[None, List[Schedule]]:
    """
    Get schedules
    dataType = fraction
    :return: None or [Schedule]
    """
    return self._schedules

  @schedules.setter
  def schedules(self, value):
    """
    Set schedules
    dataType = fraction
    :param value: [Schedule]
    """
    self._schedules = value



"""
Energy System catalog generation system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from __future__ import annotations
from typing import Union


class GenerationSystem:
  """
  Generation system class
  """
  def __init__(self, system_id, name, system_type, fuel_type, source_types, heat_efficiency, cooling_efficiency,
               electricity_efficiency, source_temperature, source_mass_flow, storage, auxiliary_equipment):

    self._system_id = system_id
    self._name = name
    self._type = system_type
    self._fuel_type = fuel_type
    self._source_types = source_types
    self._heat_efficiency = heat_efficiency
    self._cooling_efficiency = cooling_efficiency
    self._electricity_efficiency = electricity_efficiency
    self._source_temperature = source_temperature
    self._source_mass_flow = source_mass_flow
    self._storage = storage
    self._auxiliary_equipment = auxiliary_equipment

  @property
  def id(self):
    """
    Get system id
    :return: float
    """
    return self._system_id

  @property
  def name(self):
    """
    Get name
    :return: string
    """
    return self._name

  @property
  def type(self):
    """
    Get type
    :return: string
    """
    return self._type

  @property
  def fuel_type(self):
    """
    Get fuel_type from [renewable, gas, diesel, electricity, wood, coal]
    :return: string
    """
    return self._fuel_type

  @property
  def source_types(self):
    """
    Get source_type from [air, water, geothermal, district_heating, grid, on_site_electricity]
    :return: [string]
    """
    return self._source_types

  @property
  def heat_efficiency(self):
    """
    Get heat_efficiency
    :return: float
    """
    return self._heat_efficiency

  @property
  def cooling_efficiency(self):
    """
    Get cooling_efficiency
    :return: float
    """
    return self._cooling_efficiency

  @property
  def electricity_efficiency(self):
    """
    Get electricity_efficiency
    :return: float
    """
    return self._electricity_efficiency

  @property
  def source_temperature(self):
    """
    Get source_temperature in degree Celsius
    :return: float
    """
    return self._source_temperature

  @property
  def source_mass_flow(self):
    """
    Get source_mass_flow in kg/s
    :return: float
    """
    return self._source_mass_flow

  @property
  def storage(self):
    """
    Get boolean storage exists
    :return: bool
    """
    return self._storage

  @property
  def auxiliary_equipment(self) -> Union[None, GenerationSystem]:
    """
    Get auxiliary_equipment
    :return: GenerationSystem
    """
    return self._auxiliary_equipment

  def to_dictionary(self):
    """Class content to dictionary"""
    _auxiliary_equipment = []
    if self.auxiliary_equipment is not None:
      _auxiliary_equipment = self.auxiliary_equipment.to_dictionary()
    content = {'Layer': {'id': self.id,
                         'name': self.name,
                         'type': self.type,
                         'fuel type': self.fuel_type,
                         'source types': self.source_types,
                         'source temperature [Celsius]': self.source_temperature,
                         'source mass flow [kg/s]': self.source_mass_flow,
                         'heat efficiency': self.heat_efficiency,
                         'cooling efficiency': self.cooling_efficiency,
                         'electricity efficiency': self.electricity_efficiency,
                         'it has storage': self.storage,
                         'auxiliary equipment': _auxiliary_equipment
                         }
               }
    return content



"""
TestInselExports test
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import logging
from pathlib import Path
from unittest import TestCase
import pandas as pd
import hub.helpers.constants as cte
from hub.helpers.monthly_values import MonthlyValues
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.construction_factory import ConstructionFactory
from hub.imports.usage_factory import UsageFactory
from hub.exports.energy_building_exports_factory import EnergyBuildingsExportsFactory
from hub.imports.weather_factory import WeatherFactory


class TestExports(TestCase):
  """
  TestExports class contains the unittest for export functionality
  """
  def setUp(self) -> None:
    """
    Test setup
    :return: None
    """
    self._city = None
    self._complete_city = None
    self._example_path = (Path(__file__).parent / 'tests_data').resolve()
    self._output_path = (Path(__file__).parent / 'tests_outputs').resolve()

  def _get_citygml(self, file):
    file_path = (self._example_path / file).resolve()
    self._city = GeometryFactory('citygml', path=file_path).city
    self.assertIsNotNone(self._city, 'city is none')
    return self._city

  def _set_irradiance_surfaces(self, city):
    """
    saves in building surfaces the correspondent irradiance at different time-scales depending on the mode
    if building is None, it saves all buildings' surfaces in file, if building is specified, it saves only that
    specific building values
    :parameter city: city
    :return: none
    """
    city.level_of_detail.surface_radiation = 2
    path = (self._example_path / "one_building_in_kelowna_sra_SW.out").resolve()
    self._results = pd.read_csv(path, sep='\s+', header=0).to_dict(orient='list')
    _irradiance = {}
    for key in self._results:
      header_name = key.split(':')
      result = [x / cte.WATTS_HOUR_TO_JULES for x in self._results[key]]
      city_object_name = header_name[1]
      building = self._city.city_object(city_object_name)
      surface_id = header_name[2]
      surface = building.surface_by_id(surface_id)
      monthly_result = MonthlyValues.get_total_month(result)
      yearly_result = [sum(result)]
      _irradiance[cte.YEAR] = yearly_result
      _irradiance[cte.MONTH] = monthly_result
      _irradiance[cte.HOUR] = result
      surface.global_irradiance = _irradiance

  def test_insel_monthly_energy_balance_export(self):
    """
    export to Insel MonthlyEnergyBalance
    """
    city = self._get_citygml('one_building_in_kelowna.gml')
    WeatherFactory('epw', city).enrich()
    for building in city.buildings:
      building.external_temperature[cte.MONTH] = MonthlyValues().\
        get_mean_values(building.external_temperature[cte.HOUR])
    self._set_irradiance_surfaces(city)

    for building in city.buildings:
      self.assertIsNotNone(building.external_temperature[cte.MONTH], f'building {building.name} '
                                                                     f'external_temperature is none')
      for surface in building.surfaces:
        if surface.type != 'Ground':
          self.assertIsNotNone(surface.global_irradiance[cte.MONTH], f'surface in building {building.name} '
                                                                     f'global_irradiance is none')

    for building in city.buildings:
      building.year_of_construction = 2006
      if building.function is None:
        building.function = 'large office'
      building.attic_heated = 0
      building.basement_heated = 0
    ConstructionFactory('nrel', city).enrich()
    UsageFactory('comnet', city).enrich()

    # parameters written:
    for building in city.buildings:
      self.assertIsNotNone(building.volume, f'building {building.name} volume is none')
      self.assertIsNotNone(building.average_storey_height, f'building {building.name} average_storey_height is none')
      self.assertIsNotNone(building.storeys_above_ground, f'building {building.name} storeys_above_ground is none')
      self.assertIsNotNone(building.attic_heated, f'building {building.name} attic_heated is none')
      self.assertIsNotNone(building.basement_heated, f'building {building.name} basement_heated is none')
      for internal_zone in building.internal_zones:
        self.assertIsNotNone(internal_zone.area, f'internal zone {internal_zone.id} area is none')
        for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
          self.assertIsNotNone(thermal_zone.indirectly_heated_area_ratio, f'thermal zone {thermal_zone.id} '
                                                                          f'indirectly_heated_area_ratio is none')
          self.assertIsNotNone(thermal_zone.effective_thermal_capacity, f'thermal zone {thermal_zone.id} '
                                                                        f'effective_thermal_capacity is none')
          self.assertIsNotNone(thermal_zone.additional_thermal_bridge_u_value, f'thermal zone {thermal_zone.id} '
                                                                               f'additional_thermal_bridge_u_value '
                                                                               f'is none')
          self.assertIsNotNone(thermal_zone.total_floor_area, f'thermal zone {thermal_zone.id} '
                                                              f'total_floor_area is none')
          for thermal_boundary in thermal_zone.thermal_boundaries:
            self.assertIsNotNone(thermal_boundary.type)
            self.assertIsNotNone(thermal_boundary.opaque_area)
            if thermal_boundary.type in (cte.WALL, cte.ROOF):
              self.assertIsNotNone(thermal_boundary.window_ratio)
            self.assertIsNotNone(thermal_boundary.u_value)
            self.assertIsNotNone(thermal_boundary.thermal_openings)
            if thermal_boundary.type is not cte.GROUND:
              self.assertIsNotNone(thermal_boundary.external_surface.short_wave_reflectance)

        for usage in internal_zone.usages:
          self.assertIsNotNone(usage.percentage, f'usage zone {usage.name} percentage is none')
          self.assertIsNotNone(usage.internal_gains, f'usage zone {usage.name} internal_gains is none')
          self.assertIsNotNone(usage.thermal_control, f'usage zone {usage.name} thermal_control is none')
          self.assertIsNotNone(usage.hours_day, f'usage zone {usage.name} hours_day is none')
          self.assertIsNotNone(usage.days_year, f'usage zone {usage.name} days_year is none')
          self.assertIsNotNone(
            usage.mechanical_air_change,
            f'usage zone {usage.name} mechanical_air_change is none'
          )
    # export files
    try:
      EnergyBuildingsExportsFactory('insel_monthly_energy_balance', city, self._output_path, 'MEB_Montreal').export()
    except Exception as err:
      logging.exception(err)
      self.fail("Insel MonthlyEnergyBalance ExportsFactory raised ExceptionType unexpectedly!")



"""
Result factory retrieve the specific tool results and store the data in the given city
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from pathlib import Path

from hub.helpers.utils import validate_import_export_type
from hub.imports.results.insel_monthly_energry_balance import InselMonthlyEnergyBalance
from hub.imports.results.simplified_radiosity_algorithm import SimplifiedRadiosityAlgorithm


class ResultFactory:
  """
  ResultFactory class
  """

  def __init__(self, handler, city, base_path=None, hp_model=None):
    """

    :param handler: pointer to results class to be called
    :param city: the city object
    :param base_path: the path to result output file
    :param hp_model: (optional) the heat pump model for which
    results are being retrieved
    """
    if base_path is None:
      base_path = Path(Path(__file__).parent.parent / 'data/results')
    self._handler = '_' + handler.lower().replace(' ', '_')
    validate_import_export_type(ResultFactory, handler)
    self._city = city
    self._base_path = base_path
    self._hp_model = hp_model

  def _sra(self):
    """
    Enrich the city with Simplified Radiosity Algorithm results
    """
    SimplifiedRadiosityAlgorithm(self._city, self._base_path).enrich()

  def _insel_monthly_energy_balance(self):
    """
    Enrich the city with insel monthly energy balance results
    """
    InselMonthlyEnergyBalance(self._city, self._base_path).enrich()

  def enrich(self):
    """
    Enrich the city given to the class using the usage factory given handler
    :return: None
    """
    getattr(self, self._handler, lambda: None)()



"""
Greenery catalog publish the greenery information
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from pathlib import Path
from typing import TypeVar

from hub.catalog_factories.greenery.greenery_catalog import GreeneryCatalog

Catalog = TypeVar('Catalog')


class GreeneryCatalogFactory:
  """
  GreeneryCatalogFactory class
  """
  def __init__(self, handler, base_path=None):
    if base_path is None:
      base_path = (Path(__file__).parent.parent / 'data/greenery').resolve()
    self._handler = '_' + handler.lower()
    self._path = base_path

  @property
  def _nrel(self):
    """
    Return a greenery catalog based in NREL using ecore as datasource
    :return: GreeneryCatalog
    """
    return GreeneryCatalog((self._path / 'ecore_greenery_catalog.xml').resolve())

  @property
  def catalog(self) -> Catalog:
    """
    Enrich the city given to the class using the class given handler
    :return: Catalog
    """
    return getattr(self, self._handler, lambda: None)



"""
Insel monthly energy balance
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guillermo.GutierrezMorote@concordia.ca
"""

from pathlib import Path
import csv

import hub.helpers.constants as cte


class InselMonthlyEnergyBalance:
  """
  Import insel monthly energy balance results
  """
  def __init__(self, city, base_path):

    self._city = city
    self._base_path = base_path

  @staticmethod
  def _conditioning_demand(insel_output_file_path):
    heating = []
    cooling = []
    with open(Path(insel_output_file_path).resolve(), 'r', encoding='utf8') as csv_file:
      csv_reader = csv.reader(csv_file)
      for line in csv_reader:
        demand = str(line).replace("['", '').replace("']", '').split()
        for i in range(0, 2):
          if demand[i] != 'NaN':
            aux = float(demand[i]) * cte.WATTS_HOUR_TO_JULES * 1000  # kWh to J
            demand[i] = str(aux)
          else:
            demand[i] = '0'
        heating.append(float(demand[0]))
        cooling.append(float(demand[1]))
    return heating, cooling

  def _dhw_and_electric_demand(self):
    for building in self._city.buildings:
      domestic_hot_water_demand = []
      lighting_demand = []
      appliances_demand = []
      if building.internal_zones[0].thermal_zones_from_internal_zones is None:
        domestic_hot_water_demand = [0] * 12
        lighting_demand = [0] * 12
        appliances_demand = [0] * 12
      else:
        thermal_zone = building.internal_zones[0].thermal_zones_from_internal_zones[0]
        area = thermal_zone.total_floor_area
        cold_water = building.cold_water_temperature[cte.MONTH]
        peak_flow = thermal_zone.domestic_hot_water.peak_flow
        service_temperature = thermal_zone.domestic_hot_water.service_temperature
        lighting_density = thermal_zone.lighting.density
        appliances_density = thermal_zone.appliances.density

        for month in range(0, 12):
          total_dhw_demand = 0
          total_lighting = 0
          total_appliances = 0

          for schedule in thermal_zone.lighting.schedules:
            total_day = 0
            for value in schedule.values:
              total_day += value
            for day_type in schedule.day_types:
              total_lighting += total_day * cte.WEEK_DAYS_A_MONTH[day_type][month] \
                                * lighting_density / cte.WATTS_HOUR_TO_JULES
          lighting_demand.append(total_lighting * area)

          for schedule in thermal_zone.appliances.schedules:
            total_day = 0
            for value in schedule.values:
              total_day += value
            for day_type in schedule.day_types:
              total_appliances += total_day * cte.WEEK_DAYS_A_MONTH[day_type][month] \
                                  * appliances_density / cte.WATTS_HOUR_TO_JULES
          appliances_demand.append(total_appliances * area)

          for schedule in thermal_zone.domestic_hot_water.schedules:
            total_day = 0
            for value in schedule.values:
              total_day += value
            for day_type in schedule.day_types:
              demand = (
                  peak_flow * cte.WATER_DENSITY * cte.WATER_HEAT_CAPACITY
                  * (service_temperature - cold_water[month]) / cte.WATTS_HOUR_TO_JULES
              )
              total_dhw_demand += total_day * cte.WEEK_DAYS_A_MONTH[day_type][month] * demand
          domestic_hot_water_demand.append(total_dhw_demand * area)

      building.domestic_hot_water_heat_demand[cte.MONTH] = domestic_hot_water_demand
      building.domestic_hot_water_heat_demand[cte.YEAR] = [sum(domestic_hot_water_demand)]
      building.lighting_electrical_demand[cte.MONTH] = lighting_demand
      building.lighting_electrical_demand[cte.YEAR] = [sum(lighting_demand)]
      building.appliances_electrical_demand[cte.MONTH] = appliances_demand
      building.appliances_electrical_demand[cte.YEAR] = [sum(appliances_demand)]

  def enrich(self):
    """
    Enrich the city by using the insel monthly energy balance output files (J)
    :return: None
    """
    for building in self._city.buildings:
      file_name = building.name + '.out'
      insel_output_file_path = Path(self._base_path / file_name).resolve()
      if insel_output_file_path.is_file():
        building.heating_demand[cte.MONTH], building.cooling_demand[cte.MONTH] \
          = self._conditioning_demand(insel_output_file_path)
        building.heating_demand[cte.YEAR] = [sum(building.heating_demand[cte.MONTH])]
        building.cooling_demand[cte.YEAR] = [sum(building.cooling_demand[cte.MONTH])]
    self._dhw_and_electric_demand()



"""
Geojson module parses geojson files and import the geometry into the city model structure
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guillermo Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import uuid
import json

import numpy as np

from pyproj import Transformer

import hub.helpers.constants as cte
from hub.helpers.geometry_helper import GeometryHelper
from hub.imports.geometry.helpers.geometry_helper import GeometryHelper as igh
from hub.city_model_structure.attributes.polygon import Polygon
from hub.city_model_structure.building import Building
from hub.city_model_structure.building_demand.surface import Surface
from hub.city_model_structure.city import City


class Geojson:
  """
  Geojson class
  """
  _X = 0
  _Y = 1

  def __init__(self,
               path,
               aliases_field=None,
               extrusion_height_field=None,
               year_of_construction_field=None,
               function_field=None,
               function_to_hub=None):
    # todo: destination epsg should change according actual the location
    self._transformer = Transformer.from_crs('epsg:4326', 'epsg:26911')
    self._min_x = cte.MAX_FLOAT
    self._min_y = cte.MAX_FLOAT
    self._max_x = cte.MIN_FLOAT
    self._max_y = cte.MIN_FLOAT
    self._max_z = 0
    self._city = None
    self._aliases_field = aliases_field
    self._extrusion_height_field = extrusion_height_field
    self._year_of_construction_field = year_of_construction_field
    self._function_field = function_field
    self._function_to_hub = function_to_hub
    with open(path, 'r', encoding='utf8') as json_file:
      self._geojson = json.loads(json_file.read())

  def _save_bounds(self, x, y):
    if x > self._max_x:
      self._max_x = x
    if x < self._min_x:
      self._min_x = x
    if y > self._max_y:
      self._max_y = y
    if y < self._min_y:
      self._min_y = y

  @staticmethod
  def _find_wall(line_1, line_2):
    for i in range(0, 2):
      j = 1 - i
      point_1 = line_1[i]
      point_2 = line_2[j]
      distance = GeometryHelper.distance_between_points(point_1, point_2)
      if distance > 1e-2:
        return False
    return True

  def _store_shared_percentage_to_walls(self, city, city_mapped):
    for building in city.buildings:
      if building.name not in city_mapped.keys():
        for wall in building.walls:
          wall.percentage_shared = 0
        continue
      building_mapped = city_mapped[building.name]
      for wall in building.walls:
        percentage = 0
        ground_line = []
        for point in wall.perimeter_polygon.coordinates:
          if point[2] < 0.5:
            ground_line.append(point)
        for entry in building_mapped:
          if building_mapped[entry]['shared_points'] <= 2:
            continue
          line = [building_mapped[entry]['line_start'], building_mapped[entry]['line_end']]
          neighbour_line = [building_mapped[entry]['neighbour_line_start'],
                            building_mapped[entry]['neighbour_line_end']]
          neighbour_height = city.city_object(building_mapped[entry]['neighbour_name']).max_height
          if self._find_wall(line, ground_line):
            line_shared = (GeometryHelper.distance_between_points(line[0], line[1]) +
                           GeometryHelper.distance_between_points(neighbour_line[0], neighbour_line[1]) -
                           GeometryHelper.distance_between_points(line[1], neighbour_line[0]) -
                           GeometryHelper.distance_between_points(line[0], neighbour_line[1])) / 2
            percentage_ground = line_shared / GeometryHelper.distance_between_points(line[0], line[1])
            percentage_height = neighbour_height / building.max_height
            percentage_height = min(percentage_height, 1)
            percentage += percentage_ground * percentage_height
        wall.percentage_shared = percentage

  @property
  def city(self) -> City:
    """
    Get city out of a Geojson file
    """
    if self._city is None:
      buildings = []
      lod = 0
      for feature in self._geojson['features']:
        extrusion_height = 0
        if self._extrusion_height_field is not None:
          extrusion_height = float(feature['properties'][self._extrusion_height_field])
          lod = 1
        year_of_construction = None
        if self._year_of_construction_field is not None:
          year_of_construction = int(feature['properties'][self._year_of_construction_field])
        function = None
        if self._function_field is not None:
          function = str(feature['properties'][self._function_field])
          if self._function_to_hub is not None:
            # use the transformation dictionary to retrieve the proper function
            if function in self._function_to_hub:
              function = self._function_to_hub[function]
        geometry = feature['geometry']
        building_aliases = []
        if 'id' in feature:
          building_name = feature['id']
        else:
          building_name = uuid.uuid4()
        if self._aliases_field is not None:

          for alias_field in self._aliases_field:
            building_aliases.append(feature['properties'][alias_field])

        if str(geometry['type']).lower() == 'polygon':
          buildings.append(self._parse_polygon(geometry['coordinates'],
                                               building_name,
                                               building_aliases,
                                               function,
                                               year_of_construction,
                                               extrusion_height))

        elif str(geometry['type']).lower() == 'multipolygon':
          buildings.append(self._parse_multi_polygon(geometry['coordinates'],
                                                     building_name,
                                                     building_aliases,
                                                     function,
                                                     year_of_construction,
                                                     extrusion_height))
        else:
          raise NotImplementedError(f'Geojson geometry type [{geometry["type"]}] unknown')
      self._city = City([self._min_x, self._min_y, 0.0], [self._max_x, self._max_y, self._max_z], 'epsg:26911')
      for building in buildings:
        # Do not include "small building-like structures" to buildings
        if building.floor_area >= 25:
          self._city.add_city_object(building)
      self._city.level_of_detail.geometry = lod
      for building in self._city.buildings:
        building.level_of_detail.geometry = lod
      if lod > 0:
        lines_information = GeometryHelper.city_mapping(self._city, plot=False)
        self._store_shared_percentage_to_walls(self._city, lines_information)

    return self._city

  def _polygon_coordinates_to_3d(self, polygon_coordinates):
    transformed_coordinates = ''
    for coordinate in polygon_coordinates:
      transformed = self._transformer.transform(coordinate[self._Y], coordinate[self._X])
      self._save_bounds(transformed[self._X], transformed[self._Y])
      transformed_coordinates = f'{transformed_coordinates} {transformed[self._X]} {transformed[self._Y]} 0.0'
    return transformed_coordinates.lstrip(' ')

  def _parse_polygon(self, coordinates, building_name, building_aliases, function, year_of_construction, extrusion_height):
    surfaces = []
    for polygon_coordinates in coordinates:
      points = igh.points_from_string(
        igh.remove_last_point_from_string(
          self._polygon_coordinates_to_3d(polygon_coordinates)
        )
      )
      points = igh.invert_points(points)
      polygon = Polygon(points)
      polygon.area = igh.ground_area(points)
      surface = Surface(polygon, polygon)
      if surface.type == cte.GROUND:
        surfaces.append(surface)
      else:
        distance = cte.MAX_FLOAT
        hole_connect = 0
        surface_connect = 0
        for hole_index, hole_coordinate in enumerate(polygon.coordinates):
          for surface_index, ground_coordinate in enumerate(surfaces[-1].solid_polygon.coordinates):
            current_distance = GeometryHelper.distance_between_points(hole_coordinate, ground_coordinate)
            if current_distance < distance:
              distance = current_distance
              hole_connect = hole_index
              surface_connect = surface_index

        hole = polygon.coordinates[hole_connect:] + polygon.coordinates[:hole_connect] + [polygon.coordinates[hole_connect]]
        prefix_coordinates = surfaces[-1].solid_polygon.coordinates[:surface_connect+1]
        trail_coordinates = surfaces[-1].solid_polygon.coordinates[surface_connect:]
        coordinates = prefix_coordinates + hole + trail_coordinates
        polygon = Polygon(coordinates)
        polygon.area = igh.ground_area(coordinates)
        surfaces[-1] = Surface(polygon, polygon)
    if len(surfaces) > 1:
      raise ValueError('too many surfaces!!!!')
    building = Building(f'{building_name}', surfaces, year_of_construction, function)
    for alias in building_aliases:
      building.add_alias(alias)
    if extrusion_height == 0:
      return building

    volume = 0
    for ground in building.grounds:
      volume += ground.solid_polygon.area * extrusion_height
      roof_coordinates = []
      # adding a roof means invert the polygon coordinates and change the Z value
      for coordinate in ground.solid_polygon.coordinates:
        roof_coordinate = np.array([coordinate[0], coordinate[1], extrusion_height])
        # insert the roof rotated already
        roof_coordinates.insert(0, roof_coordinate)
      roof_polygon = Polygon(roof_coordinates)
      roof_polygon.area = ground.solid_polygon.area
      roof = Surface(roof_polygon, roof_polygon)
      surfaces.append(roof)
      # adding a wall means add the point coordinates and the next point coordinates with Z's height and 0
      coordinates_length = len(roof.solid_polygon.coordinates)
      for i, coordinate in enumerate(roof.solid_polygon.coordinates):
        j = i + 1
        if j == coordinates_length:
          j = 0
        next_coordinate = roof.solid_polygon.coordinates[j]
        wall_coordinates = [
          np.array([coordinate[0], coordinate[1], 0.0]),
          np.array([next_coordinate[0], next_coordinate[1], 0.0]),
          np.array([next_coordinate[0], next_coordinate[1], next_coordinate[2]]),
          np.array([coordinate[0], coordinate[1], coordinate[2]])
        ]
        polygon = Polygon(wall_coordinates)
        wall = Surface(polygon, polygon)
        surfaces.append(wall)
      building = Building(f'{building_name}', surfaces, year_of_construction, function)
      for alias in building_aliases:
        building.add_alias(alias)
      building.volume = volume
      return building

  def _parse_multi_polygon(self, polygons_coordinates, building_name, building_aliases, function, year_of_construction, extrusion_height):
    surfaces = []
    for coordinates in polygons_coordinates:
      for polygon_coordinates in coordinates:
        points = igh.points_from_string(
          igh.remove_last_point_from_string(
            self._polygon_coordinates_to_3d(polygon_coordinates)
          )
        )
        points = igh.invert_points(points)
        polygon = Polygon(points)
        polygon.area = igh.ground_area(points)
        surface = Surface(polygon, polygon)
        if surface.type == cte.GROUND:
          surfaces.append(surface)
        else:
          distance = cte.MAX_FLOAT
          hole_connect = 0
          surface_connect = 0
          for hole_index, hole_coordinate in enumerate(polygon.coordinates):
            for surface_index, ground_coordinate in enumerate(surfaces[-1].solid_polygon.coordinates):
              current_distance = GeometryHelper.distance_between_points(hole_coordinate, ground_coordinate)
              if current_distance < distance:
                distance = current_distance
                hole_connect = hole_index
                surface_connect = surface_index
          hole = polygon.coordinates[hole_connect:] + polygon.coordinates[:hole_connect]
          prefix_coordinates = surfaces[-1].solid_polygon.coordinates[:surface_connect]
          trail_coordinates = surfaces[-1].solid_polygon.coordinates[surface_connect:]
          coordinates = prefix_coordinates + hole + [hole[0]] + trail_coordinates
          polygon = Polygon(coordinates)
          polygon.area = igh.ground_area(coordinates)
          surfaces[-1] = Surface(polygon, polygon)
    building = Building(f'{building_name}', surfaces, year_of_construction, function)
    for alias in building_aliases:
      building.add_alias(alias)
    if extrusion_height == 0:
      return building

    volume = 0
    for ground in building.grounds:
      volume += ground.solid_polygon.area * extrusion_height
      roof_coordinates = []
      # adding a roof means invert the polygon coordinates and change the Z value
      for coordinate in ground.solid_polygon.coordinates:
        roof_coordinate = np.array([coordinate[0], coordinate[1], extrusion_height])
        # insert the roof rotated already
        roof_coordinates.insert(0, roof_coordinate)
      roof_polygon = Polygon(roof_coordinates)
      roof_polygon.area = ground.solid_polygon.area
      roof = Surface(roof_polygon, roof_polygon)
      surfaces.append(roof)
      # adding a wall means add the point coordinates and the next point coordinates with Z's height and 0
      coordinates_length = len(roof.solid_polygon.coordinates)
      for i, coordinate in enumerate(roof.solid_polygon.coordinates):
        j = i + 1
        if j == coordinates_length:
          j = 0
        next_coordinate = roof.solid_polygon.coordinates[j]
        wall_coordinates = [
          np.array([coordinate[0], coordinate[1], 0.0]),
          np.array([next_coordinate[0], next_coordinate[1], 0.0]),
          np.array([next_coordinate[0], next_coordinate[1], next_coordinate[2]]),
          np.array([coordinate[0], coordinate[1], coordinate[2]])
        ]
        polygon = Polygon(wall_coordinates)
        wall = Surface(polygon, polygon)
        surfaces.append(wall)
      building = Building(f'{building_name}', surfaces, year_of_construction, function)
      for alias in building_aliases:
        building.add_alias(alias)
      building.volume = volume
    return building



"""
TestExports test and validate the city export formats
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import logging.handlers
from pathlib import Path
from unittest import TestCase
from hub.imports.geometry_factory import GeometryFactory
from hub.helpers.dictionaries import Dictionaries
from hub.imports.construction_factory import ConstructionFactory
from hub.imports.usage_factory import UsageFactory
from hub.imports.weather_factory import WeatherFactory
from hub.exports.exports_factory import ExportsFactory
from hub.exports.energy_building_exports_factory import EnergyBuildingsExportsFactory
import hub.helpers.constants as cte
from hub.city_model_structure.city import City


class TestExports(TestCase):
  """
  TestExports class contains the unittest for export functionality
  """
  def setUp(self) -> None:
    """
    Test setup
    :return: None
    """
    self._city = None
    self._complete_city = None
    self._example_path = (Path(__file__).parent / 'tests_data').resolve()
    self._output_path = (Path(__file__).parent / 'tests_outputs').resolve()

  def _get_citygml(self, file):
    file_path = (self._example_path / file).resolve()
    self._city = GeometryFactory('citygml', path=file_path).city
    self.assertIsNotNone(self._city, 'city is none')
    return self._city

  def _get_complete_city(self, from_pickle):
    if self._complete_city is None:
      if from_pickle:
        file_path = (self._example_path / 'ConcordiaSWGcampus.pickle').resolve()
        self._complete_city = City.load(file_path)
      else:
        file_path = (self._example_path / 'one_building_in_kelowna.gml').resolve()
        self._complete_city = self._get_citygml(file_path)
        for building in self._complete_city.buildings:
          building.function = Dictionaries().hft_function_to_hub_function[building.function]
          building.year_of_construction = 2006
        ConstructionFactory('nrel', self._complete_city).enrich()
        UsageFactory('nrcan', self._complete_city).enrich()
        cli = (self._example_path / 'weather' / 'inseldb_Summerland.cli').resolve()
        self._complete_city.climate_file = Path(cli)
        self._complete_city.climate_reference_city = 'Summerland'
        dummy_measures = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        for building in self._complete_city.buildings:
          building.heating_demand[cte.MONTH] = dummy_measures
          building.cooling_demand[cte.MONTH] = dummy_measures
          building.heating_demand[cte.YEAR] = [0.0]
          building.cooling_demand[cte.YEAR] = [0.0]
    return self._complete_city

  def _export(self, export_type, from_pickle=False):
    self._complete_city = self._get_complete_city(from_pickle)
    try:
      ExportsFactory(export_type, self._complete_city, self._output_path).export()
    except ValueError as err:
      if export_type != 'stl':
        logging.warning('No backend export for STL test, skipped')
        raise err

  def _export_building_energy(self, export_type, from_pickle=False):
      self._complete_city = self._get_complete_city(from_pickle)
      EnergyBuildingsExportsFactory(export_type, self._complete_city, self._output_path).export()

  def test_obj_export(self):
    """
    export to obj
    """
    self._export('obj', False)

  def test_stl_export(self):
    """
    export to stl
    """
    self._export('stl', False)

  def test_energy_ade_export(self):
    """
    export to energy ADE
    """
    self._export_building_energy('energy_ade')

  def test_sra_export(self):
    """
    export to SRA
    """
    self._export('sra')

  def test_idf_export(self):
    """
    export to IDF
    """
    file = 'test.geojson'
    file_path = (self._example_path / file).resolve()
    city = GeometryFactory('geojson',
                           path=file_path,
                           height_field='citygml_me',
                           year_of_construction_field='ANNEE_CONS',
                           function_field='CODE_UTILI',
                           function_to_hub=Dictionaries().montreal_function_to_hub_function).city

    self.assertIsNotNone(city, 'city is none')
    EnergyBuildingsExportsFactory('idf', city, self._output_path).export()
    ConstructionFactory('nrcan', city).enrich()
    EnergyBuildingsExportsFactory('idf', city, self._output_path).export()
    UsageFactory('nrcan', city).enrich()
    WeatherFactory('epw', city).enrich()
    print(self._output_path)
    try:
      EnergyBuildingsExportsFactory('idf', city, self._output_path).export()
    except Exception:
      self.fail("Idf ExportsFactory raised ExceptionType unexpectedly!")




"""
Node module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
import uuid
from typing import List, TypeVar

Node = TypeVar('Node')


class Edge:
  """
  Generic edge class to be used as base for the network edges
  """
  def __init__(self, name, nodes=None):
    if nodes is None:
      nodes = []
    self._name = name
    self._id = None
    self._nodes = nodes

  @property
  def name(self):
    """
    Get edge name
    :return: str
    """
    return self._name

  @property
  def id(self):
    """
    Get edge id, a universally unique identifier randomly generated
    :return: str
    """
    if self._id is None:
      self._id = uuid.uuid4()
    return self._id

  @property
  def nodes(self) -> List[Node]:
    """
    Get delimiting nodes for the edge
    :return: [Node]
    """
    return self._nodes



"""
Base repository class to establish db connection
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
import logging
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from hub.persistence.configuration import Configuration


class Repository:
  """
  Base repository class to establish db connection
  """

  def __init__(self, db_name, dotenv_path: str, app_env='TEST'):
    try:
      self.configuration = Configuration(db_name, dotenv_path, app_env)
      self.engine = create_engine(self.configuration.connection_string)
      self.session = Session(self.engine)
    except ValueError as err:
      logging.error('Missing value for credentials: %s', err)



"""
Archetype catalog Cost
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from hub.catalog_factories.data_models.cost.capital_cost import CapitalCost
from hub.catalog_factories.data_models.cost.operational_cost import OperationalCost
from hub.catalog_factories.data_models.cost.income import Income


class Archetype:
  """
  Archetype class
  """
  def __init__(self,
               lod,
               function,
               municipality,
               country,
               currency,
               capital_cost,
               operational_cost,
               end_of_life_cost,
               income):

    self._lod = lod
    self._function = function
    self._municipality = municipality
    self._country = country
    self._currency = currency
    self._capital_cost = capital_cost
    self._operational_cost = operational_cost
    self._end_of_life_cost = end_of_life_cost
    self._income = income

  @property
  def name(self):
    """
    Get name
    :return: string
    """
    return f'{self._country}_{self._municipality}_{self._function}_lod{self._lod}'

  @property
  def lod(self):
    """
    Get level of detail of the catalog
    :return: string
    """
    return self._lod

  @property
  def function(self):
    """
    Get function
    :return: string
    """
    return self._function

  @property
  def municipality(self):
    """
    Get municipality
    :return: string
    """
    return self._municipality

  @property
  def country(self):
    """
    Get country
    :return: string
    """
    return self._country

  @property
  def currency(self):
    """
    Get currency
    :return: string
    """
    return self._currency

  @property
  def capital_cost(self) -> CapitalCost:
    """
    Get capital cost
    :return: CapitalCost
    """
    return self._capital_cost

  @property
  def operational_cost(self) -> OperationalCost:
    """
    Get operational cost
    :return: OperationalCost
    """
    return self._operational_cost

  @property
  def end_of_life_cost(self):
    """
    Get end of life cost in given currency per m2
    :return: float
    """
    return self._end_of_life_cost

  @property
  def income(self) -> Income:
    """
    Get income
    :return: Income
    """
    return self._income

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {'Archetype': {'name': self.name,
                             'level of detail': self.lod,
                             'municipality': self.municipality,
                             'country': self.country,
                             'currency': self.currency,
                             'function': self.function,
                             'capital cost': self.capital_cost.to_dictionary(),
                             'operational cost': self.operational_cost.to_dictionary(),
                             'end of life cost [currency/m2]': self.end_of_life_cost,
                             'income': self.income.to_dictionary()
                             }
               }
    return content



"""
Generic energy generation system definition
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from __future__ import annotations
from typing import Union


class GenericGenerationSystem:
  """
  GenericGenerationSystem class
  """
  def __init__(self):
    self._type = None
    self._fuel_type = None
    self._source_types = None
    self._heat_efficiency = None
    self._cooling_efficiency = None
    self._electricity_efficiency = None
    self._source_temperature = None
    self._source_mass_flow = None
    self._storage = None
    self._auxiliary_equipment = None

  @property
  def type(self):
    """
    Get system type
    :return: string
    """
    return self._type

  @type.setter
  def type(self, value):
    """
    Set system type
    :param value: string
    """
    self._type = value

  @property
  def fuel_type(self):
    """
    Get fuel_type from [Renewable, Gas, Diesel, Electricity, Wood, Coal]
    :return: string
    """
    return self._fuel_type

  @fuel_type.setter
  def fuel_type(self, value):
    """
    Set fuel_type from [Renewable, Gas, Diesel, Electricity, Wood, Coal]
    :param value: string
    """
    self._fuel_type = value

  @property
  def source_types(self):
    """
    Get source_type from [Air, Water, Geothermal, District Heating, Grid, Onsite Electricity]
    :return: [string]
    """
    return self._source_types

  @source_types.setter
  def source_types(self, value):
    """
    Set source_type from [Air, Water, Geothermal, District Heating, Grid, Onsite Electricity]
    :param value: [string]
    """
    self._source_types = value

  @property
  def heat_efficiency(self):
    """
    Get heat_efficiency
    :return: float
    """
    return self._heat_efficiency

  @heat_efficiency.setter
  def heat_efficiency(self, value):
    """
    Set heat_efficiency
    :param value: float
    """
    self._heat_efficiency = value

  @property
  def cooling_efficiency(self):
    """
    Get cooling_efficiency
    :return: float
    """
    return self._cooling_efficiency

  @cooling_efficiency.setter
  def cooling_efficiency(self, value):
    """
    Set cooling_efficiency
    :param value: float
    """
    self._cooling_efficiency = value

  @property
  def electricity_efficiency(self):
    """
    Get electricity_efficiency
    :return: float
    """
    return self._electricity_efficiency

  @electricity_efficiency.setter
  def electricity_efficiency(self, value):
    """
    Set electricity_efficiency
    :param value: float
    """
    self._electricity_efficiency = value

  @property
  def source_temperature(self):
    """
    Get source_temperature in degree Celsius
    :return: float
    """
    return self._source_temperature

  @source_temperature.setter
  def source_temperature(self, value):
    """
    Set source_temperature in degree Celsius
    :param value: float
    """
    self._source_temperature = value

  @property
  def source_mass_flow(self):
    """
    Get source_mass_flow in kg/s
    :return: float
    """
    return self._source_mass_flow

  @source_mass_flow.setter
  def source_mass_flow(self, value):
    """
    Set source_mass_flow in kg/s
    :param value: float
    """
    self._source_mass_flow = value

  @property
  def storage(self):
    """
    Get boolean storage exists
    :return: bool
    """
    return self._storage

  @storage.setter
  def storage(self, value):
    """
    Set boolean storage exists
    :return: bool
    """
    self._storage = value

  @property
  def auxiliary_equipment(self) -> Union[None, GenericGenerationSystem]:
    """
    Get auxiliary_equipment
    :return: GenerationSystem
    """
    return self._auxiliary_equipment

  @auxiliary_equipment.setter
  def auxiliary_equipment(self, value):
    """
    Set auxiliary_equipment
    :return: GenerationSystem
    """
    self._auxiliary_equipment = value



"""
Greenery catalog data model Soil class
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""


class Soil:
  """
  Soil class
  """
  def __init__(self, soil):
    self._name = soil.name
    self._roughness = soil.roughness
    self._dry_conductivity = soil.conductivityOfDrySoil
    self._dry_density = soil.densityOfDrySoil
    self._dry_specific_heat = soil.specificHeatOfDrySoil
    self._thermal_absorptance = soil.thermalAbsorptance
    self._solar_absorptance = soil.solarAbsorptance
    self._visible_absorptance = soil.visibleAbsorptance
    self._saturation_volumetric_moisture_content = soil.saturationVolumetricMoistureContent
    self._residual_volumetric_moisture_content = soil.residualVolumetricMoistureContent
    self._initial_volumetric_moisture_content = soil.initialVolumetricMoistureContent

  @property
  def name(self):
    """
    Get soil name
    :return: string
    """
    return self._name

  @property
  def roughness(self):
    """
    Get soil roughness
    :return: string
    """
    return self._roughness

  @property
  def dry_conductivity(self):
    """
    Get soil dry conductivity in W/mK
    :return: float
    """
    return self._dry_conductivity

  @property
  def dry_density(self):
    """
    Get soil dry density in kg/m3
    :return: float
    """
    return self._dry_density

  @property
  def dry_specific_heat(self):
    """
    Get soil dry specific heat in J/kgK
    :return: float
    """
    return self._dry_specific_heat

  @property
  def thermal_absorptance(self):
    """
    Get soil thermal absortance
    :return: float
    """
    return self._thermal_absorptance

  @property
  def solar_absorptance(self):
    """
    Get soil solar absortance
    :return: float
    """
    return self._solar_absorptance

  @property
  def visible_absorptance(self):
    """
    Get soil visible absortance
    :return: float
    """
    return self._visible_absorptance

  @property
  def saturation_volumetric_moisture_content(self):
    """
    Get soil saturation volumetric moisture content
    :return: float
    """
    return self._saturation_volumetric_moisture_content

  @property
  def residual_volumetric_moisture_content(self):
    """
    Get soil residual volumetric moisture content
    :return: float
    """
    return self._residual_volumetric_moisture_content

  @property
  def initial_volumetric_moisture_content(self):
    """
    Get soil initial volumetric moisture content
    :return: float
    """
    return self._initial_volumetric_moisture_content

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {'Soil': {'name': self.name,
#                        'roughness': self.roughness,   # todo: this line prints value=2????
                        'dry conductivity [W/m2K]': self.dry_conductivity,
                        'dry density [kg/m3]': self.dry_density,
                        'dry specific heat [J/kgK]': self.dry_specific_heat,
                        'thermal absorptance': self.thermal_absorptance,
                        'solar absorptance': self.solar_absorptance,
                        'visible absorptance': self.visible_absorptance,
                        'saturation volumetric moisture content [units??]': self.saturation_volumetric_moisture_content,
                        'residual volumetric moisture content [units??]': self.residual_volumetric_moisture_content
                        }
               }

    return content



"""
Layers module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

import uuid
from typing import Union


class Layer:
  """
  Layer class
  """
  def __init__(self):
    self._thickness = None
    self._id = None
    self._name = None
    self._conductivity = None
    self._specific_heat = None
    self._density = None
    self._solar_absorptance = None
    self._thermal_absorptance = None
    self._visible_absorptance = None
    self._no_mass = False
    self._thermal_resistance = None

  @property
  def id(self):
    """
    Get layer id, a universally unique identifier randomly generated
    :return: str
    """
    if self._id is None:
      self._id = uuid.uuid4()
    return self._id

  @property
  def thickness(self) -> Union[None, float]:
    """
    Get layer thickness in meters
    :return: None or float
    """
    return self._thickness

  @thickness.setter
  def thickness(self, value):
    """
    Get layer thickness in meters
    :param value: float
    """
    if value is not None:
      self._thickness = float(value)

  @property
  def name(self):
    """
    Get material name
    :return: str
    """
    return self._name

  @name.setter
  def name(self, value):
    """
    Set material name
    :param value: string
    """
    self._name = str(value)

  @property
  def conductivity(self) -> Union[None, float]:
    """
    Get material conductivity in W/mK
    :return: None or float
    """
    return self._conductivity

  @conductivity.setter
  def conductivity(self, value):
    """
    Set material conductivity in W/mK
    :param value: float
    """
    if value is not None:
      self._conductivity = float(value)

  @property
  def specific_heat(self) -> Union[None, float]:
    """
    Get material conductivity in J/kgK
    :return: None or float
    """
    return self._specific_heat

  @specific_heat.setter
  def specific_heat(self, value):
    """
    Get material conductivity in J/kgK
    :param value: float
    """
    if value is not None:
      self._specific_heat = float(value)

  @property
  def density(self) -> Union[None, float]:
    """
    Get material density in kg/m3
    :return: None or float
    """
    return self._density

  @density.setter
  def density(self, value):
    """
    Set material density
    :param value: float
    """
    if value is not None:
      self._density = float(value)

  @property
  def solar_absorptance(self) -> Union[None, float]:
    """
    Get material solar absorptance
    :return: None or float
    """
    return self._solar_absorptance

  @solar_absorptance.setter
  def solar_absorptance(self, value):
    """
    Set material solar absorptance
    :param value: float
    """
    if value is not None:
      self._solar_absorptance = float(value)

  @property
  def thermal_absorptance(self) -> Union[None, float]:
    """
    Get material thermal absorptance
    :return: None or float
    """
    return self._thermal_absorptance

  @thermal_absorptance.setter
  def thermal_absorptance(self, value):
    """
    Set material thermal absorptance
    :param value: float
    """
    if value is not None:
      self._thermal_absorptance = float(value)

  @property
  def visible_absorptance(self) -> Union[None, float]:
    """
    Get material visible absorptance
    :return: None or float
    """
    return self._visible_absorptance

  @visible_absorptance.setter
  def visible_absorptance(self, value):
    """
    Set material visible absorptance
    :param value: float
    """
    if value is not None:
      self._visible_absorptance = float(value)

  @property
  def no_mass(self) -> Union[None, bool]:
    """
    Get material no mass flag
    :return: None or Boolean
    """
    return self._no_mass

  @no_mass.setter
  def no_mass(self, value):
    """
    Set material no mass flag
    :param value: Boolean
    """
    if value is not None:
      self._no_mass = value

  @property
  def thermal_resistance(self) -> Union[None, float]:
    """
    Get material thermal resistance in m2K/W
    :return: None or float
    """
    return self._thermal_resistance

  @thermal_resistance.setter
  def thermal_resistance(self, value):
    """
    Set material thermal resistance in m2K/W
    :param value: float
    """
    if value is not None:
      self._thermal_resistance = float(value)



"""
Generic energy distribution system definition
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""


class GenericDistributionSystem:
  """
  GenericDistributionSystem class
  """
  def __init__(self):
    self._type = None
    self._supply_temperature = None
    self._distribution_consumption_fix_flow = None
    self._distribution_consumption_variable_flow = None
    self._heat_losses = None

  @property
  def type(self):
    """
    Get type from [air, water, refrigerant]
    :return: string
    """
    return self._type

  @type.setter
  def type(self, value):
    """
    Set type from [air, water, refrigerant]
    :param value: string
    """
    self._type = value

  @property
  def supply_temperature(self):
    """
    Get supply_temperature in degree Celsius
    :return: float
    """
    return self._supply_temperature

  @supply_temperature.setter
  def supply_temperature(self, value):
    """
    Set supply_temperature in degree Celsius
    :param value: float
    """
    self._supply_temperature = value

  @property
  def distribution_consumption_fix_flow(self):
    """
    Get distribution_consumption if the pump or fan work at fix mass or volume flow in ratio over peak power (W/W)
    :return: float
    """
    return self._distribution_consumption_fix_flow

  @distribution_consumption_fix_flow.setter
  def distribution_consumption_fix_flow(self, value):
    """
    Set distribution_consumption if the pump or fan work at fix mass or volume flow in ratio over peak power (W/W)
    :return: float
    """
    self._distribution_consumption_fix_flow = value

  @property
  def distribution_consumption_variable_flow(self):
    """
    Get distribution_consumption if the pump or fan work at variable mass or volume flow in ratio
    over energy produced (J/J)
    :return: float
    """
    return self._distribution_consumption_variable_flow

  @distribution_consumption_variable_flow.setter
  def distribution_consumption_variable_flow(self, value):
    """
    Set distribution_consumption if the pump or fan work at variable mass or volume flow in ratio
    over energy produced (J/J)
    :return: float
    """
    self._distribution_consumption_variable_flow = value

  @property
  def heat_losses(self):
    """
    Get heat_losses in ratio over energy produced
    :return: float
    """
    return self._heat_losses

  @heat_losses.setter
  def heat_losses(self, value):
    """
    Set heat_losses in ratio over energy produced
    :param value: float
    """
    self._heat_losses = value



"""
CityGmlBase module abstract class to template the different level of details
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from abc import ABC


class CityGmlBase(ABC):
  """
  CityGmlBase class inherited by the specific level of detail classes.
  """
  def __init__(self):
    self._surfaces = []

  @property
  def surfaces(self):
    """
    Get parsed surfaces
    """
    return self._surfaces

  @classmethod
  def _solid(cls, city_object_member):
    raise NotImplementedError

  @classmethod
  def _multi_surface(cls, city_object_member):
    raise NotImplementedError

  @classmethod
  def _multi_curve(cls, city_object_member):
    raise NotImplementedError
